# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cambrionix smart hub module.

This module contains APIs for interfacing with a Cambrionix smart hub device.

The following Cambrionix devices are currently supported::

    PP8S - A commercial quality 8 port smart hub device
    PP15S - A commercial quality 15 port smart hub device
    U16S - An uncased 16 port smart hub device

PP8S - Also known as the PowerPad 8S has 8 ports diagrammed below::

    XX      xxx
    Switch  Power
    xxx Host Port    --- Empty
    --- Empty        --- Empty
    --- Empty        --- Empty
    --- Empty        --- Empty
    xxx Port 1       xxx Port 5
    xxx Port 2       xxx Port 6
    xxx Port 3       xxx Port 7
    xxx Port 4       xxx Port 8

PP15S - Also known as the PowerPad 15S has 15 ports diagrammed below::

    XX      xxx
    Switch  Power

    xxx Host Port    xxx Port 8
    xxx Port 1       xxx Port 9
    xxx Port 2       xxx Port 10
    xxx Port 3       xxx Port 11
    xxx Port 4       xxx Port 12
    xxx Port 5       xxx Port 13
    xxx Port 6       xxx Port 14
    xxx Port 7       xxx Port 15

U16S - An uncased 16 port OEM device::

    xxx     xxx     xxx     xxx
    Port 1  Port 2  Port 3  Port 4                     xxx Host Port

    xxx     xxx     xxx     xxx                         X GND
    Port 5  Port 6  Port 7  Port 8                      X Power
                                                        X GPIO Pins?
    xxx     xxx     xxx     xxx                         X
    Port 9  Port 10 Port 11 Port 12                     X

    xxx     xxx     xxx     xxx         xxx          xxx
    Port 13 Port 14 Port 15 Port 16   Expansion 1  Expansion 2

Each of the above devices offers a serial port that can be used to control
the Cambrionix device. The code below accesses this serial port to allow
for controlling the device. Access the serial port using the following
serial port parameters::

    Baud Rate: 115200
    Data Bits: 8
       Parity: None
    Stop Bits: 1

Typical commands::

    help   - Display all available commands
    health - Display the health of the Cambrionix device
    l      - Live view of all active ports (also allows simple toggling of ports)
    mode   - Allows changing mode of a particular port (o-off, c-charge, s-sync)
    state  - Display the state of a particular port
    system - Display system information of the Cambrionix device

"""
import fcntl
import os
import select
import time

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import switch_power_usb_with_charge
from gazoo_device import detect_criteria

from gazoo_device.utility import usb_config
from gazoo_device.utility import usb_utils
import serial

logger = gdm_logger.get_gdm_logger()

COMMANDS = {
    "CLEAR_ERROR_FLAGS": "cef",
    "CLEAR_REBOOTED_FLAG": "crf",
    "LIMITS": "limits",
    "STATE": "state",
    "GET_MODE": "state {}",
    "SYSTEM_STATUS": "system",
    "REBOOT_WATCHDOG": "reboot watchdog",
    "REBOOT": "reboot",
    "SET_MODE": "mode {} {}"
}

REGEXES = {
    "GET_MODE_REGEX": r"\s*\d{1,2},\s+\w+,\s+([A-Z0-9\s]+),",
}

TIMEOUTS = {
    "OPEN": 30,  # Waits for other interactions. (reboot with watchdog=True takes ~16
    # seconds and is likely the longest running command)
    "PING": 3,
    "REBOOT": 3,
    "REBOOT_WATCHDOG": 15
}

_REBOOT_METHODS = ["watchdog", "shell"]


class Cambrionix(auxiliary_device.AuxiliaryDevice):
    """This class serves as a Python interface to a Cambrionix hub.

    Attributes:
      Hub Device Base Class
    """
    COMMUNICATION_TYPE = "SerialComms"
    DETECT_MATCH_CRITERIA = {detect_criteria.SERIAL_QUERY.product_name: "ft230x basic uart"}
    DEVICE_TYPE = "cambrionix"
    _instances = {}

    def __new__(cls, manager, device_config, log_file_name, log_directory):
        """Returns the same object for the same serial_port_path.

        Args:
            manager(gazoo_device.Manager): manager object for this instance of gazoo_device.
            device_config(dict): dict of two dicts, 'optional' device parameters and
                                  'persistent' device parameters.
            log_file_name(str): file name in the log directory for device logs.
            log_directory(str): directory in which the controller will create the log file.

        Returns:
          Cambrionix: Instance matching serial port path

        Note:
          Instantiate a new serial_port_path if not found.
        """
        # slowly migrate away from using 'hub_port_name' but maintain backwards compatibility
        if 'console_port_name' not in device_config['persistent']:
            device_config['persistent']['console_port_name'] = \
                device_config['persistent']['hub_port_name']

        identifier = device_config['persistent']['console_port_name']
        if identifier not in cls._instances:
            obj = super(Cambrionix, cls).__new__(cls)
            cls._instances[identifier] = obj

        return cls._instances[identifier]

    def __init__(self, manager, device_config, log_file_name, log_directory):
        """Constructor of the class Cambrionix.

        Args:
            manager(gazoo_device.Manager): manager object for this instance of gazoo_device.
            device_config(dict): dict of two dicts, 'optional' device parameters and
                                  'persistent' device parameters.
            log_file_name(str): file name in the log directory for device logs.
            log_directory(str): directory in which the controller will create the log file.

        Raises:
          GazooDeviceError: unknown type of cambrionix
        """
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)
        self._serial_port = None

    def __del__(self):
        self.close()

    @decorators.health_check
    def clear_flags(self):
        """clear the error and reboot flags."""
        self._command(self.commands["CLEAR_ERROR_FLAGS"])
        self._command(self.commands["CLEAR_REBOOTED_FLAG"])

    @decorators.DynamicProperty
    def firmware_version(self):
        """Gets the firmware version of the hub.

        Returns:
          str: The firmware version
        """
        return self._get_system_status()['firmware']

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected,
                self.clear_flags]

    @decorators.PersistentProperty
    def port_extension_map(self):
        """Map of usb port extensions to port numbers."""
        return usb_config.CAMBRIONIX_PORT_MAP[self.model]

    @decorators.PersistentProperty
    def total_ports(self):
        """Number of usb ports for this device."""
        return len(self.port_extension_map.keys())

    @decorators.LogDecorator(logger, level=decorators.DEBUG)
    def close(self):
        """Closes the serial port connection."""
        if self._serial_port is not None and self._serial_port.is_open:
            self._serial_port.close()

        super(Cambrionix, self).close()

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a Cambrionix.

        Returns:
          tuple: (Dict of persistent attributes, dict of optional attributes)

        Raises:
            GazooDeviceError: if device model is not supported.

        Notes:
          persistent: model,
                      hub_port_name,
                      console_port_name,
                      total_ports,
                      ftdi_serial_number,
                      serial_number

          optional:   empty dict
        """
        persistent_dict = self.props['persistent_identifiers']
        persistent_dict['model'] = self._get_system_hardware()
        if persistent_dict['model'] not in usb_config.CAMBRIONIX_PORT_MAP:
            raise errors.GazooDeviceError("Model {} not supported. Supported models: {}".
                                          format(persistent_dict['model'],
                                                 ",".join(usb_config.CAMBRIONIX_PORT_MAP.keys())))
        persistent_dict['hub_port_name'] = self.communication_address
        persistent_dict['console_port_name'] = self.communication_address
        persistent_dict['total_ports'] = self.total_ports
        persistent_dict['ftdi_serial_number'] = usb_utils.get_serial_number_from_path(
            self.communication_address)

        # Cambrionix does not have a separate serial number from the one shown
        # in the /dev/serial/by-id/... name.
        persistent_dict['serial_number'] = self.props[
            'persistent_identifiers']['ftdi_serial_number']

        self.props['options'] = {}

        return persistent_dict, self.props['options']

    @classmethod
    def is_connected(cls, device_config):
        """Returns true if a Cambrionix is connected to the computer, False otherwise.

        Args:
          device_config (dict): contains "persistent" dict

        Returns:
          bool: Cambrionix is connected

        Notes:
          To return True, the cambrionix_config needs to have "console_port_name" in its
          "persistent" dict and that path needs to exist on the computer.
        """
        if "console_port_name" in device_config["persistent"]:
            address = device_config["persistent"]["console_port_name"]
        else:
            address = device_config["persistent"]["hub_port_name"]
        return os.path.exists(address)

    @decorators.LogDecorator(logger)
    def reboot(self, no_wait=False, method="shell"):
        """Reboots the system.

        Args:
          no_wait (bool): Return before boot up is complete. Default: False.
          method (str): Method to invoke reboot. 'shell' | 'watchdog'

        Raises:
          ValueError: if method name not recognized.

        Note:
          (Terminal Command Reference v1.5 page 15).
        """
        if method.lower() not in _REBOOT_METHODS:
            raise ValueError(
                f"Method {method!r} not recognized. Supported methods: {_REBOOT_METHODS}")
        if method == 'watchdog':
            self._command(self.commands["REBOOT_WATCHDOG"],
                          close_delay=self.timeouts["REBOOT_WATCHDOG"])
        else:
            self._command(self.commands["REBOOT"], close_delay=self.timeouts["REBOOT"])

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Attempts to recover device based on the type of error specified.

        Note: The check_device_ready method can raise a number of separate exceptions which
        are passed to this method as exception objects. The recovery method is chosen based on
        the type of the error. A list of errors and recovery steps follow (not all recovery
        steps have been implemented yet):

            DeviceNotResponsiveError:
                * Recovery: reboot device.

        Args:
            error (CheckDeviceReadyError): A subclass of CheckDeviceReadyError that will be
                used to identify a possible recovery solution to use.

        Raises:
            GazooDeviceError: If device recovery fails while attempting to perform recovery
                steps.
            CheckDeviceReadyError: If there are no recovery steps available for the error
                argument, it will be re-raised directly.
        """

        if isinstance(error, errors.DeviceNotResponsiveError):
            self.reboot()
        else:
            raise error

    def shell(self,
              command,
              command_name="shell",
              timeout=None,
              port=0,
              include_return_code=False):
        """Sends command and returns response.

        Args:
            command (str): Command to send to the device.
            command_name (str): Optional identifier to use in logs for this command.
            timeout (float): Seconds to wait for pattern after command sent.
            port (int): not used.
            include_return_code (bool): flag indicating return code should be returned

        Returns:
            str or tuple: response or (response, return_code) if include_return_code is True.
                          return_code 0 - success, return_code -1 - error.
        """
        timeout = timeout or self.timeouts["SHELL"]
        response = ""
        if include_return_code:
            return_code = 0
            try:
                response = self._command(command)
                response = self._list_to_str(response)
            except errors.GazooDeviceError:
                return_code = -1
            return response, return_code
        response = self._command(command)
        return self._list_to_str(response)

    def shell_with_regex(self,
                         command,
                         regex,
                         regex_group=1,
                         command_name="shell",
                         raise_error=False,
                         tries=1,
                         port=0):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command (str): command to issue.
            regex (str): regular expression with one or more capturing groups.
            regex_group (int): number of regex group to return.
            command_name (str): command name to appear in log messages.
            raise_error (bool): whether or not to raise error if unable to find a match.
            tries (int): how many times to try executing the command before failing.
            port (int): which port to send the shell command to.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        return self.command_with_regex(command,
                                       regex,
                                       self.shell,
                                       regex_group=regex_group,
                                       raise_error=raise_error,
                                       tries=tries,
                                       command_name=command_name,
                                       port=port)

    @decorators.CapabilityDecorator(switch_power_usb_with_charge.SwitchPowerUsbWithCharge)
    def switch_power(self):
        """Returns a switch_power object to send commands.

        Returns:
            object: switch_power_usb_with_charge
        """
        return self.lazy_init(
            switch_power_usb_with_charge.SwitchPowerUsbWithCharge,
            shell_fn=self.shell,
            regex_shell_fn=self.shell_with_regex,
            command_dict=self.commands,
            regex_dict=self.regexes,
            device_name=self.name,
            serial_number=self.serial_number,
            total_ports=self.total_ports)

    @decorators.PersistentProperty
    def valid_modes(self):
        return ['off', 'sync', 'charge']

    def _command(self, command, close_delay=0.0):
        """Opens the control serial port and sends a command.

        Args:
          command (str): Command to send to device
          close_delay (float): Seconds to delay between command and closing control port

        Returns:
          list: All strings in the response (except the trailing >> prompt)

        Raises:
          GazooDeviceError: Error in response to command.

        Note:
          Returns response. Delays closing the port afterwards if close_delay > 0.

          With some commands (e.g. reboot), it is necessary to wait before closing
          the control serial port to prevent other GDM instances from accessing the
          control serial port.
        """

        self._open()
        try:
            self.__write_command(self._serial_port, command)
            if command.startswith('reboot'):
                return

            response = self.__get_response(self._serial_port)
            if response[0].startswith('*E'):
                raise errors.GazooDeviceError("Device {} command failed. "
                                              "Unable to write command: {} "
                                              "to serial port: {}  Err: {!r}".
                                              format(self.name,
                                                     command,
                                                     self._serial_port,
                                                     response[0]))
        finally:
            if close_delay > 0.0:
                time.sleep(close_delay)
            self.close()

        # Discard the last line which is the prompt
        return response[:-1]

    def _get_system_status(self):
        """Gets hardware and firmware information.

        Returns:
          dict: Information regarding the system

        Note:
          (Terminal Command Reference v1.5 page 6)
          Example output from PP8S device:
          >> system
          cambrionix PP8S 8 Port USB Charge+Sync
          Hardware: PP8S
          Firmware: 1.68
          Compiled: Feb 14 2017 17:30:26
          Group: -
          Panel ID: Absent
        """

        sysinfo_strings = self._command(self.commands["SYSTEM_STATUS"])
        sysinfo_dict = {'name': sysinfo_strings[0]}
        for line in sysinfo_strings:
            if ":" in line:
                key, value = line.split(':', 1)
                sysinfo_dict[key.lower()] = value.strip()

        return sysinfo_dict

    def _get_system_hardware(self):
        """Gets the hardware description of the hub.

        Returns:
          str: Hardware description
        """

        return self._get_system_status()['hardware']

    def _list_to_str(self, lst):
        if isinstance(lst, list):
            lst = ' '.join(map(str, lst))
        return lst

    def _open(self):
        start_time = time.time()
        error = ""
        while time.time() - start_time < self.timeouts["OPEN"]:
            try:
                if self._serial_port is None:
                    self._serial_port = serial.Serial(port=self.communication_address,
                                                      baudrate=115200,
                                                      timeout=0.1,
                                                      exclusive=True)
                    # Prevent inheritance of file descriptors to exec'd child processes [NEP-1852]
                    file_descriptor = self._serial_port.fd
                    flags = fcntl.fcntl(file_descriptor, fcntl.F_GETFD)
                    fcntl.fcntl(file_descriptor, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
                    return
                if not self._serial_port.is_open:
                    self._serial_port.open()
                return
            except Exception as err:
                error = err
        raise errors.GazooDeviceError("Device {} open failed. "
                                      "Unable to open control serial port in {} seconds"
                                      "Error: {}".
                                      format(self.name,
                                             self.timeouts["OPEN"],
                                             error))

    @staticmethod
    def __write_command(serial_port, command):
        """Internal helper for writing to the hub.

        Args:
          serial_port (str): Cambrionix serial port.
          command (str): command to send to device.

        Raises:
          GazooDeviceError: no response from device.
        """

        line_ending = '\r\n'
        ctrl_c_cmd = '\x03' + line_ending

        # Clear any existing text by sending a CTRL-C
        # command and waiting for a prompt
        serial_port.write(ctrl_c_cmd.encode('utf-8'))
        Cambrionix.__get_response(serial_port)

        if not command.endswith(line_ending):
            command += line_ending

        for char in command:
            serial_port.write(char.encode('utf-8'))
            if command.startswith("reboot") and char == '\r':
                break

            while True:
                ready = select.select([serial_port], [], [], 25)[0]
                if ready:
                    if serial_port.read(1).decode('utf-8') == char:
                        break
                else:
                    raise errors.GazooDeviceError("Device cambrionix write command failed. "
                                                  "Read timeout on serial port: {} "
                                                  "while writing command: {}".
                                                  format(serial_port,
                                                         command))

    @staticmethod
    def __get_response(serial_port):
        """Internal helper returning the response from the hub as a list of lines.

        Args:
          serial_port (str): Cambrionix serial port.

        Returns:
          str: response.

        Raises:
          GazooDeviceError: Device not responding.

        Note:
          The command prompt is always this string: ">> ".
        """

        read_data = ''
        while not read_data.endswith('\n>> '):
            ready = select.select([serial_port], [], [], 25)[0]
            if ready:
                read_data += serial_port.read(serial_port.inWaiting()).decode('utf-8', 'replace')
            else:
                raise errors.GazooDeviceError("Device cambrionix get response failed. "
                                              "Read timeout on serial port: {}".
                                              format(serial_port))

        return read_data.splitlines()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sideload capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class SideloadBase(capability_base.CapabilityBase):
    """Abstract base class for the sideload capability."""

    @abc.abstractmethod
    def enter(self):
        """Enters sideload mode for updating device firmware.

        Raises:
            GazooDeviceError: if unable to boot device into sideload mode.
        """

    @abc.abstractmethod
    def exit(self):
        """Exits sideload mode after updating device firmware.

        Raises:
            GazooDeviceError if unable to reboot device out of sideload mode
        """

    @abc.abstractmethod
    def is_in_sideload_mode(self):
        """Whether the device is currently in sideload mode.

        Returns:
            bool: True if device in sideload mode; False otherwise.
        """



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal Decorators used in GDM device classes and capabilities.

Included are the following:

* Log Decorator
* CapabilityLogDecorator
* CapabilityDecorator
* health_check
* DynamicProperty
* PersistentProperty
* SettableProperty


- BASIC USAGE:

***CapabilityDecorator***
Required for all capability properties in device classes.
Used to dynamically identify the capabilities and their flavors of a device class.

@decorators.CapabilityDecorator(<capability_flavor>)

Ex:

@decorators.CapabilityDecorator(light_events_default.LightEventsDefault)
def light_events(self):
    return self.lazy_init(light_events_default.LightEventsDefault,
                          shell_fn=self.shell,
                          device_name=self.name,
                          command_dict=self.commands)

Can declared multiple flavors, if the flavor is not determined until initiation
@decorators.CapabilityDecorator(
        [weave_default.WeaveDefault, weave_log_events_based.WeaveLogEventsBased,
         weave_shell_based.WeaveShellBased])
    def weave(self):
        ....


***LogDecorator***
Required for all public methods in a device lcass that don't return a value.

This decorator will:
    1. Log a start message at the start of decorated method
    2. Log a success message at the end of decorated method if method succeeds
        2.1 The success message will also contain the method execution time.
    3. Log a skip message if method is skipped
        3.1 To skip a method, raise SkipExceptionError("YOUR_REASON.")
    4. Wrap all your exceptions in GazooDeviceError (or a different wrap_type).
        4.1 The original exception type will be preserved in the error message.
    5. Add "<device> <method> failed. <Exception type>: <Exception message>."
        formatting to all exceptions.
        5.1 Note: don't specify the device name and method name in the exception message.
            This decorator does this for you!
            Example: use SomeError("Foo") instead of SomeError("Device <..> failed. Reason: Foo")

@decorators.LogDecorator(logger)
def your_method(self):
    pass

OR (to change the level)
Log_LEVEL can be one of NONE, DEBUG, INFO, WARNING, ERROR, CRITICAL
@decorators.LogDecorator(logger, level=decorators.<LOG_LEVEL>)
def your_method(self):
    pass


***CapabilityLogDecorator***
Required for all public methods in capability classes that don't return a value.
logger is the gdm_logger instance in the class

@decorators.CapabilityLogDecorator(logger)
def your_method(self):
    pass

OR (to change the level)
Log_LEVEL can be one of NONE, DEBUG, INFO, WARNING, ERROR, CRITICAL
@decorators.CapabilityLogDecorator(logger, level=decorators.<LOG_LEVEL>)
def your_method(self):
    pass

***health_check***
Used to identify all the public methods in a device class used for health checks.
Note: The only public methods allowed in device classes are those identified in the base template
 and health checks. All others must be in capabilities.

@decorators.health_check
def check_logs_streaming(self):

***DynamicProperty***
Used to identify dynamic properties, usually device states.

These properties change throughout testing. ex: firmware_version, pairing.state, etc.

@decorators.DynamicProperty
def firmware_version(self):

***PersistentProperty***
Used to identify all persistent properties.

These are properties that are persistent throughout device lifetime.
ex: hardware_model, serial_number, etc. These are either populated during device detection or
static for a device class.

@decorators.PersistentProperty
def serial_number(self):

***SettableProperty***
These are settable properties, usually initialized as None. They allow users to enable additional
capabilities or set properties undetectable via device communication.
Ex: power_switch_port to enable power cycling via powerswitch, alias, or device_usb_port if the
device is connected to a Cambrionix.

@decorators.SettableProperty
def alias(self):
"""
import functools
import inspect
import logging
import time
from gazoo_device import errors
from gazoo_device import gdm_logger

GDM_LOGGER = gdm_logger.get_gdm_logger()

# Enable specifying logger levels by decorators.<LEVEL> (so users don't have to import logging)
NONE = None
DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

MESSAGES = {
    "START": "{device_name} starting {class_name}.{method_name}",
    "SKIP": "{device_name} {class_name}.{method_name} skipped. {skip_reason}.",
    "SUCCESS": "{device_name} {class_name}.{method_name} successful. It took {time_elapsed}s.",
    "FAILURE": "{device_name} {class_name}.{method_name} failed. {exc_name}: {exc_reason}"
}

DEFAULT_DEVICE_NAME = "Unknown_device"

# This attribute is added to decorated methods so it's possible
# to access the wrapped function. Same as "__wrapped__" attribute in Python 3.
WRAPS_ATTR_NAME = "_gdm_decorators_wraps"


class SkipExceptionError(Exception):
    """Used to tell info Error wrapper to skip last message."""


def unwrap(func):
    """Get the original function object of a wrapper function.

    Noop (returns func) if func is not decorated with decorator.

    Args:
        func (function): method, or property decorated with decorator

    Returns:
        function: the original function object (prior to decoration).
    """
    while hasattr(func, "__wrapped__"):
        func = func.__wrapped__
    return func


class CapabilityDecorator():
    """Decorator to define capabilities in device classes (wraps CapabilityProperty).

    @decorators.CapabilityDecorator(light_events_default.LightEventsDefault)
    def light_events(self):
        return self.lazy_init(light_events_default.LightEventsDefault,
                              shell_fn=self.shell,
                              device_name=self.name,
                              command_dict=self.commands)
    """

    def __init__(self, class_or_classes):
        """Initialize the capability decorator.

        Args:
            class_or_classes (object): single capability flavor class or list (or tuple) of
                                       capability flavor classes.
                                       GDM uses this to determine which capabilities are
                                       supported by a device class.

        Note:
            Input validation of class_or_classes is not done by the decorator at import time.
            Instead, it is performed by unit_tests/test_all_devices.py.
            This allows to avoid a circular import.
        """
        classes = class_or_classes
        if not isinstance(class_or_classes, (list, tuple)):
            classes = [classes]
        self._capability_classes = classes

    def __call__(self, fget):
        """Wrapper around CapabilityProperty to support decorator syntax.

        Args:
            fget (method): getter function (returns a capability instance).
                           Note that most capability getter functions should use
                           GazooDeviceBase.lazy_init().

        Returns:
            CapabilityProperty: property denoting a supported capability.
        """
        return CapabilityProperty(fget, classes=self._capability_classes)


class CapabilityProperty(property):
    """Property under which a capability is accessible in a device class. Do not use directly.

    Disambiguates device property definitions and device capability definitions.
    """

    def __init__(self, fget, classes):
        """Initialize the capability property.

        Args:
            fget (method): getter function (returns a capability instance).
                           Note that most capability getter functions should use
                           GazooDeviceBase.lazy_init().
            classes (sequence): list or tuple of capability flavor classes.
                                Used by GDM to determine which capabilities are supported by a
                                device class.
        """
        super().__init__(fget)
        self.capability_classes = set(classes)


class LogDecorator():
    """Wraps GDM methods with standard logger messages (logged to the provided logger).

    Catches all exceptions and wraps them in GazooDeviceError (or other wrap_type).
    Required for any public method that doesn't return a value.

    Note:
        GDM design asks for there to be a syntax for all info messsages and errors raised.
        This enforces that syntax.

        Example output when successful:
        cambrionix-1234 started factory_reset
        cambrionix-1234 successfully factory_reset in 10 s.

        Example out when failed:
        cambrionix-1234 started factory_reset
        cambrionix-1234 factory_reset failed. GazooDeviceError: device did not respond
            recognizably.
    """

    def __init__(self, logger, level=INFO, wrap_type=errors.GazooDeviceError, name_attr="name"):
        """Create a log decorator wrapper.

        Args:
            logger (logger): logger to use to print messages.
            level (int or None): logging level (see logging module level for integer values).
                                 (decorators.INFO, decorators.DEBUG, decorators.NONE)
                                 NONE disables logging messages (but errors are still wrapped).
            wrap_type (GazooDeviceError): wrap the errors in this class if it's not an instance of
                                         it already. Valid values are GazooDeviceError
                                         or its subclasses.
            name_attr (str): name of attribute containing the device name.
                             For example, if device name is under self.device_name,
                             the value should be "device_name".

        Raises:
            TypeError: if wrap_type is not a subclass of GazooDeviceError.
        """
        if not issubclass(wrap_type, errors.GazooDeviceError):
            raise TypeError("Invalid wrap_type: {}. "
                            "wrap_type can only be a subclass of GazooDeviceError."
                            .format(wrap_type))
        self.logger = logger
        self.level = level
        self.wrap_type = wrap_type
        self.name_attr = name_attr

    def __call__(self, func):
        """Wraps (decorates) the provided function.

        Args:
            func (function): function to be decorated

        Returns:
            function: a wrapper function (decorated input function)

        Raises:
            TypeError: incorrect type of func argument.
        """
        # Note: cannot use inspect.ismethod() here. At the time of decorator __init__ call,
        # python intepreter has not yet created the class of the method we're decorating.
        # At this time, the 'method' is still a function. Once all methods are created,
        # the python interpreter creates the class object, then binds the functions' self
        # arguments to make them 'methods.' Hence at this point func is considered a function
        # by python even if it's actually a method.
        if not callable(func):
            raise TypeError("Expected func to be callable, found {}.".format(func))

        error_template = ("Expected {} to be an instance method. "
                          "Decorating {} methods is not supported; remove the decorator.")
        func_args = inspect.getfullargspec(func).args

        if func_args and func_args[0] == "cls":
            raise TypeError(error_template.format(func, "class"))
        if not func_args or func_args[0] != "self":
            raise TypeError(error_template.format(func, "static"))

        @functools.wraps(func)
        def wrapped_func(instance, *args, **kwargs):
            """Wraps (decorates) the given function.

            Args:
                instance (object): class instance
                *args (tuple): positional arguments to the wrapped function
                **kwargs (dict): keyword arguments to the wrapped function

            Raises:
                TracebackError: wrapped error (the traceback is added to __str__ and __repr__).

            Note:
                this wrapper only works for methods, not functions (see the instance argument).

            Returns:
                object: same value as the wrapped function.
            """
            fmt_args = {
                "device_name": getattr(instance, self.name_attr, DEFAULT_DEVICE_NAME),
                "method_name": func.__name__,
                "class_name": self._find_defining_class_name(func, type(instance)),
                "skip_reason": None,
                "time_elapsed": None,
                "exc_name": None,
                "exc_reason": None
            }

            return_val = None
            method_skipped = False
            start_time = time.time()

            if self.level is not None:
                self.logger.log(self.level, MESSAGES["START"].format(**fmt_args))

            try:
                return_val = func(instance, *args, **kwargs)
            except SkipExceptionError as err:
                method_skipped = True
                fmt_args["skip_reason"] = str(err)
            except Exception as err:
                self._format_and_raise(fmt_args, err)

            if self.level is not None:
                if method_skipped:
                    self.logger.log(self.level, MESSAGES["SKIP"].format(**fmt_args))
                else:
                    fmt_args["time_elapsed"] = int(time.time() - start_time)
                    self.logger.log(self.level, MESSAGES["SUCCESS"].format(**fmt_args))

            return return_val

        wrapped_func.__dict__[WRAPS_ATTR_NAME] = func
        return wrapped_func

    def _find_defining_class_name(self, method, current_class):
        """Finds the name of the class from which the method was inherited from.

        Args:
            method (func): method object.
            current_class (type): class to start the search from.

        Returns:
            str: name of the class defining the given method.
            None: defining class wasn't found in the class hierarchy.

        Note:
            There are 2 subtleties:
            * to correctly identify the class for super().<method> calls, the method check has to
              compare the method by identity (to avoid matching overridden methods).
            * methods may need to be unwrapped (from the log decorator) to make the identity
              comparison work.
            Doesn't work on class methods, but they are not decorated with the log decorator.
        """
        for a_class in current_class.__mro__:
            a_class_method = vars(a_class).get(method.__name__)
            if a_class_method and unwrap(a_class_method) is method:
                return a_class.__name__
        return None

    def _format_and_raise(self, fmt_args, err):
        """Format the exception message & type before re-raising."""
        wrapped_exc = err
        if (not isinstance(err, errors.CheckDeviceReadyError)
                and not isinstance(err, self.wrap_type)):
            fmt_args["exc_name"] = type(err).__name__
            fmt_args["exc_reason"] = str(err)
            # Wrap the error in a different type and reraise.
            reraise_msg = MESSAGES["FAILURE"].format(**fmt_args)
            wrapped_exc = self.wrap_type(reraise_msg)

        # Note: do nothing for CheckDeviceReadyErrors & subclasses.
        # These have two mandatory positional arguments in their signatures,
        # unlike the other errors, which typically have just one (the message).
        # It would be nice to get rid of the extra arguments (device name) in the signatures
        # and instead just let the log decorator add the device name to the error message.
        # However, refactoring these errors to use the log decorator is non-trivial.
        # This is because many of these are raised in plain functions (not methods),
        # which are not supported by this log decorator...
        raise wrapped_exc


class CapabilityLogDecorator(LogDecorator):
    """Adds standardized log messages to capability methods."""

    def __init__(self, logger, level=INFO, wrap_type=errors.GazooDeviceError,
                 name_attr="_device_name"):
        """Log decorator for capabilities (with a different name_attr default value).

        Args:
            logger (logger): logger to use to print messages.
            level (int or None): logging level (see logging module level for integer values).
                                 For example, level=INFO or level=NONE.
                                 NONE disables logging messages (but errors are still wrapped).
            wrap_type (GazooDeviceError): wrap the errors in this class if it's not an instance of
                                         it already. Valid values are GazooDeviceError
                                         or its subclasses.
            name_attr (str): name of attribute containing the device name.
                             For example, if device name is under self.device_name,
                             the value should be "device_name".

        Note:
            All capabilities need to have a device name reference in the capability class.
            This is required for log and error messages in multi-device tests.
        """
        super().__init__(logger, level=level, wrap_type=wrap_type, name_attr=name_attr)


decorators_factory = LogDecorator(GDM_LOGGER, level=DEBUG)


def health_check(func):
    """Health Check Decorator.  Returns silenced log decorator."""
    func.__health_check__ = True
    return decorators_factory(func)


class DynamicProperty(property):
    """A property that is dynamic and involves a device query to return."""

    def __init__(self, fget):
        super().__init__(fget, doc=fget.__doc__)
        self.name = fget.__name__


class SettableProperty(property):
    """Usually defaults to null, a property that can be set."""

    def __init__(self, fget, fset=None, fdel=None, doc=None):
        if not doc:
            doc = fget.__doc__
        super().__init__(fget, fset=fset, fdel=fdel, doc=doc)
        self.name = fget.__name__


class PersistentProperty(property):
    """A property that is persistent throughout device interaction.

    Is static across device class or can only be reset during device detection.
    """

    def __init__(self, fget):
        super().__init__(fget, doc=fget.__doc__)
        self.name = fget.__name__
